{
  Merge Plugins Script v1.9.0
  제작: matortheeternal
  번역: renn
  http://skyrim.nexusmods.com/mod/37981
  
  *설명*
  이 스크립트를 통해 esp 파일을 병합할 수 있습니다. 깨진 데이터가 있으면 병합이 취소됩니다.
  병합될 파일 선택시 고급 옵션(나사 아이콘) 설정창에서 원하는 옵션대로 설정할 수 있습니다.
  
  *업데이트 요약*
  v1.9.0
  - (MO 전용) Asset (플러그인 이름이 경로에 포함된 파일) 저장 경로 도우미 버튼 이름 변경
  - 중요한 충돌 FormID 오류가 로그에 기록됨
  - Injected FormID 리넘버링 개정됨 (병합된 Injected FormID의 복제 오류 방지)
  - (MO 전용) Asset 복사시 파일 전체 검색에서 모드 폴더 내 검색 방식으로 변경됨
    esp/esm이 해당 모드 폴더 안에 없을 경우 검색 취소됨
  - Scripts\mp\temp 폴더를 임시 폴더로 사용, 확인된 파일은 임시 폴더에 안전하게 저장됨
  - cmd 배치 복사시 병합 로그에 기록됨
  - 병합 로그 이름이 제대로 생성됨
  - Fixed CopyGeneralAssets scanning "." and ".." folders on certain systems.
  - Implemented user-friendly EditOutOfDate method from mteFunctions when it has
    been detected that the user is using an out of date version of xEdit.
  - Now using robocopy instead of the deprecated xcopy function when copying
    general assets with batch copying.	
  - Now using copy instead of robocopy when useRobocopy constant is false.  This
    fix should help users who have privilege escalation issues when starting xEdit
    through Mod Organizer.
  - Updated EditOutOfDate method to take a URL as a parameter, now pointing to Nexus
    Mods for TES5Edit (because of the 3.1.0 release).  
}

unit mergePlugins;

uses mteFunctions;

const
  vs = 'v1.9';
  dashes = '-----------------------------------------------------------------------------------';
  debug = false; // debug messages
  debugMCM = false;
  debugRenumbering = false;
  debugAssetCopying = false;
  debugSearch = false;
  useRobocopy = false;
  deleteTemp = true;
  pFlag = 'Record Header\Record Flags\PersistentReference QuestItem DisplaysInMainMenu';

var
  slMerge, slMasters, slFails, slSelectedFiles, slMgfMasters, slDictionary, 
  slTranslations, slCopiedFrom, batch: TStringList;
  OldForms, NewForms: TList;
  rn, mm, sp, rCount: integer;
  moPath, astPath, bsaName, temp, fdt: string;
  SkipProcess, disableColoring, extractBSAs, usingMo, copyAll, 
  firstRun, batCopy: boolean;
  mgf: IInterface;
  cbArray: Array[0..254] of TCheckBox;
  lbArray: Array[0..254] of TLabel;
  slArray: Array[0..30] of TStringList;
  frm: TForm;
  memo: TMemo;
  mnPopup: TPopupMenu;
  btnDetails, btnSave: TButton;
  gear, browse: TPicture;
  cb1, cb2: TCheckbox;
  btnFind: TButton;
  ed1, ed2: TEdit;
  imgBrowse1, imgBrowse2, imgBrowse3: TImage;
  NextObjectID: cardinal;
  Records: array [0..$FFFFFF] of IInterface;
  UsedFormIDs: array [0..$FFFFFF] of byte;
 
 
{*************************************************************************}
{***************************** GUI Functions *****************************}
{*************************************************************************}

//=========================================================================
// LogMessage: Posts a message to the log stringlist
procedure LogMessage(msg: String);
begin
  memo.Lines.add(msg);
  Application.processmessages; // doesn't appear to cause issues
end;
  
//=========================================================================
// GetDefinitionHint: Generates a hint based on the definition
function GetDefinitionHint(sl: TStringList): string;
var
  notes: String;
begin
  if sl.Count < 6 then
    Result := '이 플러그인에 제출된 사용자 리포트가 없습니다.'
  else begin
    notes := Trim(StringReplace(sl[5], '@13', #13, [rfReplaceAll]));
    Result := '평균 평가: '+sl[3]+#13+'평가 수: '+sl[4]+#13+'사용자 노트: '+#13+notes;
  end;
end;
  
//=========================================================================
// GetMergeColor: gets the color associated with the file's merge rating
function GetMergeColor(sl: TStringList): integer;
var
  rating, k1, k2: float;
  c1, c2, c3, fc: TColor;
begin
  if sl.Count < 2 then
    Result := $404040
  else begin
    c1.Red := $FF; c1.Green := $00; c1.Blue := $00;
    c2.Red := $E5; c2.Green := $A8; c2.Blue := $00;
    c3.Red := $00; c3.Green := $90; c3.Blue := $00;
    fc.Blue := $00;
    rating := StrToFloat(sl[3]);
    if (rating > 2.0) then begin
      k2 := (rating - 2.0)/2.0;
      k1 := 1.0 - k2;
      fc.Red := c2.Red * k1 + c3.Red * k2;
      fc.Green := c2.Green * k1 + c3.Green * k2;
    end
    else begin
      k2 := (rating/2.0);
      k1 := 1.0 - k2;
      fc.Red := c1.Red * k1 + c2.Red * k2;
      fc.Green := c1.Green * k1 + c2.Green * k2;
    end;
    Result := ColorToInt(fc.Red, fc.Green, fc.Blue);
  end;
end;

//=========================================================================
// GetDefinition: gets a definition for the file from the dictionary
function GetDefinition(f: IInterface; r: boolean; v: boolean): string;
var
  i: integer;
  search: string;
begin
  Result := '';
  search := GetFileName(f) + ';';
  if r then
    search := search + IntToStr(RecordCount(f)) + ';';
  if v then
    search := search + vs + ';';
  for i := 0 to Pred(slDictionary.Count) do begin
    if Pos(search, slDictionary[i]) = 1 then begin
      Result := slDictionary[i];
      break;
    end;
  end;
end;

//=========================================================================
// GetAllDefinitions: gets all definition for the file from the dictionary
function GetAllDefinitions(f: IInterface): string;
var
  i: integer;
  search: string;
begin
  Result := '';
  search := GetFileName(f) + ';';
  for i := 0 to Pred(slDictionary.Count) do begin
    if Pos(search, slDictionary[i]) = 1 then
      Result := Result + slDictionary[i] + ';';
  end;
end;

//=========================================================================
// FileSelectM: File selection window for merging
function FileSelectM(lbl: string): IInterface;
var
  cmbFiles: TComboBox;
  btnOk, btnCancel: TButton;
  lbl01: TLabel;
  i, j, llo: integer;
  s: string;
  f: IInterface;
begin
  frm := TForm.Create(frm);
  try
    frm.Caption := '파일 선택';
    frm.Width := 290;
    frm.Height := 230;
    frm.Position := poScreenCenter;
    
    lbl01 := TLabel.Create(frm);
    lbl01.Parent := frm;
    lbl01.Width := 250;
    lbl01.Height := 60;
    lbl01.Left := 8;
    lbl01.Top := 8;
    lbl01.Caption := lbl;
    lbl01.Autosize := false;
    lbl01.Wordwrap := True;
    
    cmbFiles := TComboBox.Create(frm);
    cmbFiles.Parent := frm;
    cmbFiles.Items.Add('-- 새 파일로 생성 --');
    cmbFiles.Style := csDropDownList;
    cmbFiles.Top := 33 + lbl01.Height;
    cmbFiles.Left := 8;
    cmbFiles.Width := 225;
    llo := 0;
    
    for j := 0 to slMerge.Count - 1 do 
      if llo < Integer(slMerge.Objects[j]) then 
        llo := Integer(slMerge.Objects[j]);
    for i := 0 to FileCount - 1 do begin
      s := GetFileName(FileByIndex(i));
      if Pos(s, bethesdaFiles) > 0 then Continue;
      if slMerge.IndexOf(s) > -1 then Continue;
      if GetLoadOrder(FileByIndex(i)) < llo then Continue;
      cmbFiles.Items.Add(s);
    end;
    cmbFiles.ItemIndex := 0;
    
    ConstructOkCancelButtons(frm, frm, cmbFiles.Top + 50);
    
    if frm.ShowModal = mrOk then begin
      if (cmbFiles.Items[cmbFiles.ItemIndex] = '-- 새 파일로 생성 --') then begin
        f := AddNewFile;
        Result := f;
      end
      else begin
        for i := 0 to FileCount - 1 do begin
          if (cmbFiles.Items[cmbFiles.ItemIndex] = GetFileName(FileByIndex(i))) then begin
            Result := FileByIndex(i);
            Continue;
          end;
          if i = FileCount - 1 then begin
            AddMessage('입력한 파일을 스크립트가 찾지 못했습니다.');
            Result := FileSelectM(lbl);
          end;
        end;
      end;
    end;
  finally
    frm.Free;
  end;
end;

//=========================================================================
// CheckDirectories: Checks if ed1 and ed2 directories are valid.
procedure ofrm.CheckDirectories;
var
  moValid, astValid: boolean;
begin
  moValid := DirectoryExists(ed1.Caption);
  astValid := DirectoryExists(ed2.Caption);
  btnSave.Enabled := (moValid or not cb1.Checked) and astValid;
end;

//=========================================================================
// MoPathHelper: Tells the user if their Mod Organizer path is invalid
procedure ofrm.MoPathHelper;
var
  moValid: boolean;
begin
  moValid := DirectoryExists(ed1.Caption);
  if not moValid then begin
    // tell user their mod organizer path is invalid
    MessageDlg('Mod Organizer 경로가 유효하지 않습니다. 올바른 경로를 '
      '입력하거나 "Mod Organizer 사용자" 옵션의 체크를 해제하세요.', mtConfirmation, [mbOk], 0);
  end;
end;

//=========================================================================
// AssetPathHelper: Tells the user if their asset path is invalid
procedure ofrm.AssetPathHelper;
var
  astValid: boolean;
begin
  astValid := DirectoryExists(ed2.Caption);
  if not astValid then begin
    // tell user their asset is invalid
    MessageDlg('Asset가 복사될 폴더 경로가 유효하지 않습니다. '
      '올바른 경로를 입력하세요.', mtConfirmation, [mbOk], 0);
  end;
end;

//=========================================================================
// AssetPathBrowse: Browse for asset destination path
procedure ofrm.AssetPathBrowse;
var
  s: string;
begin
  if DirectoryExists(ed2.Caption) then
    s := SelectDirectory('폴더 선택', '', ed2.Text, '')
  else if cb1.Checked and DirectoryExists(ed1.Caption) then
    s := SelectDirectory('폴더 선택', '', ed1.Text, '')
  else
    s := SelectDirectory('폴더 선택', '', DataPath, '');
  if s <> '' then begin
    ed2.Text := s + '\';
  end;
end;

//=========================================================================
// MoPathBrowse: Browse for Mod Organizer path
procedure ofrm.MoPathBrowse;
var
  s: string;
begin
  s := SelectDirectory('폴더 선택', '', ed1.Text, '');
  if s <> '' then begin
    ed1.Text := s + '\';
  end;
end;

//=========================================================================
// UsingModOrganizer: Toggle for controls
procedure ofrm.UsingModOrganizer;
begin
  ed1.Enabled := cb1.Checked;
  btnFind.Enabled := cb1.Checked;
  cb2.Enabled := cb1.Checked;
  imgBrowse1.Enabled := cb1.Checked;
  CheckDirectories;
end;

//=========================================================================
// DetectModOrganizer: searches for ModOrganizer.exe
procedure ofrm.DetectModOrganizer;
var
  i: int;
  modOrganizerPath, paths, v: string;
  pathList, ignore: TStringList;
  rec: TSearchRec;
begin
  // search for installations in ?:\Program Files and ?:\Program Files (x86)
  for i := 65 to 90 do begin
    if DirectoryExists(chr(i) + ':\Program Files') then
      paths := paths + chr(i) + ':\Program Files;';
    if DirectoryExists(chr(i) + ':\Program Files (x86)') then
      paths := paths + chr(i) + ':\Program Files (x86);';
  end;
  
  modOrganizerPath := FileSearch('Mod Organizer\ModOrganizer.exe', paths);
  
  // search for installations in GamePath
  if (modOrganizerPath = '') then begin
    ignore := TStringList.Create;
    ignore.Add('data');
    modOrganizerPath := RecursiveFileSearch('ModOrganizer.exe', GamePath, ignore, 2, debugSearch);
  end;
  
  // search each folder in each valid Program Files directory for ModOrganizer.exe
  if (modOrganizerPath = '') then begin
    pathList := TStringList.Create;
    while (Pos(';', paths) > 0) do begin
      pathList.Add(Copy(paths, 1, Pos(';', paths) - 1));
      paths := Copy(paths, Pos(';', paths) + 1, Length(paths));
    end;
    for i := 0 to pathList.Count - 1 do begin
      if FindFirst(pathList[i] + '\*', faDirectory, rec) = 0 then begin
        repeat
          if debugSearch then AddMessage('검색 중 '+pathList[i]+'\'+rec.Name);
          modOrganizerPath := FileSearch('ModOrganizer.exe', pathList[i] + '\' + rec.Name);
          if (modOrganizerPath <> '') then begin
            //modOrganizerPath := Copy(modOrganizerPath, 1, Length(modOrganizerPath) - 1);
            break;
          end;
        until FindNext(rec) <> 0;
        
        FindClose(rec);
        if (modOrganizerPath <> '') then break;
      end;
    end;
  end;
  
  // if found, set TEdit captions, else alert user
  if (modOrganizerPath <> '') then begin
    ed1.Caption := Copy(modOrganizerPath, 1, length(modOrganizerPath) - 16);
    ed2.Caption := ed1.Caption + 'overwrite\';
  end
  else begin
    AddMessage('Mod Organizer의 경로를 자동으로 탐색할 수 없습니다. 수동으로 직접 입력해 주세요.');
    ed1.Caption := '';
  end;
end;

//=========================================================================
// SaveSettings: Saves the current script settings/options
procedure SaveSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ScriptsPath + 'mp\config.ini');
  ini.WriteBool('Config', 'usingMO', usingMo);
  ini.WriteString('Config', 'moPath', moPath);
  ini.WriteString('Config', 'astPath', astPath);
  ini.WriteBool('Config', 'copyAllAssets', copyAll);
  ini.WriteString('Config', 'renumberingMode', IntToStr(rn));
  ini.WriteString('Config', 'copyMode', IntToStr(mm));
  ini.WriteString('Config', 'secondPassMode', IntToStr(sp));
  ini.WriteBool('Config', 'disableColoring', disableColoring);
  ini.WriteBool('Config', 'extractBSAs', extractBSAs);
  ini.WriteBool('Config', 'batCopy', batCopy);
  ini.UpdateFile;
end;

//=========================================================================
// SaveSettings: Saves the current script settings/options
procedure LoadSettings;
var
  ini: TMemIniFile;
  cfgPath: string;
begin
  cfgPath := FileSearch('mp\config.ini', ScriptsPath);
  if (cfgPath = '') then begin
    MessageDlg('플러그인 병합 스크립트를 처음 실행하셨습니다. '
      '나사 아이콘을 눌러 고급 옵션창을 열고 병합 옵션을 설정하세요. ', mtConfirmation, [mbOk], 0);
    firstRun := true;
  end;
  ini := TMemIniFile.Create(ScriptsPath + 'mp\config.ini');
  usingMO := ini.ReadBool('Config', 'usingMO', false);
  moPath := ini.ReadString('Config', 'moPath', '');
  astPath := ini.ReadString('Config', 'astPath', DataPath);
  copyAll := ini.ReadBool('Config', 'copyAllAssets', false);
  rn := IntToStr(ini.ReadString('Config', 'renumberingMode', '1'));
  mm := IntToStr(ini.ReadString('Config', 'copyMode', '1'));
  sp := IntToStr(ini.ReadString('Config', 'secondPassMode', '0'));
  disableColoring := ini.ReadBool('Config', 'disableColoring', false);
  extractBSAs := ini.ReadBool('Config', 'extractBSAs', false);
  batCopy := ini.ReadBool('Config', 'batCopy', true);
end;

//=========================================================================
// AdvancedOptions: 
procedure AdvancedOptions;
var
  ofrm: TForm;
  lbl1, lbl2: TLabel;
  cb3, cb4, cb5: TCheckBox;
  gb1, gb2: TGroupBox;
  btnDiscard: TButton;
  rg1, rg2, rg3: TRadioGroup;
  rb1, rb2, rb3, rb4, rb5, rb6, rb7, rb8, rb9: TRadioButton;
begin
  ofrm := TForm.Create(nil);
  try
    ofrm.Caption := '고급 옵션';
    ofrm.Width := 610;
    ofrm.Position := poScreenCenter;
    ofrm.Height := 550;
    
    gb1 := TGroupBox.Create(ofrm);
    gb1.Parent := ofrm;
    gb1.Left := 16;
    gb1.Height := 120;
    gb1.Top := 16;
    gb1.Width := 560;
    gb1.Caption := 'Mod Organizer 옵션';
    gb1.ClientHeight := 105;
    gb1.ClientWidth := 556;
    
    cb1 := TCheckBox.Create(gb1);
    cb1.Parent := gb1;
    cb1.Left := 16;
    cb1.Top := 20;
    cb1.Width := 150;
    cb1.Caption := ' Mod Organizer 사용자   ';
    cb1.Checked := usingMO;
    cb1.OnClick := UsingModOrganizer;
    cb1.Hint := 
      'Mod Organizer 사용자의 경우, '#13
	  'Asset 복사를 제대로 하려면 이 옵션에 체크해야 합니다. '#13
      'Mod Organizer 사용자가 아닌 경우에는 체크해제하세요.';
    cb1.ShowHint := true;
    
    lbl1 := TLabel.Create(gb1);
    lbl1.Parent := gb1;
    lbl1.Left := 16;
    lbl1.Top := cb1.Top + cb1.Height + 12;
    lbl1.Width := 90;
    lbl1.Caption := 'Mod Organizer 설치 경로: ';
    
    ed1 := TEdit.Create(gb1);
    ed1.Parent := gb1;
    ed1.Left := lbl1.Left + lbl1.Width + 24;
    ed1.Top := lbl1.Top;
    ed1.Width := 250;
    ed1.Caption := moPath;
    ed1.Enabled := usingMO;
    ed1.Hint := 
      'ModOrganizer.exe가 들어 있는 폴더의 경로입니다. '#13
	  '"Mod Organizer 사용자" 옵션에 체크하셨다면 반드시 설정하세요.';
    ed1.ShowHint := true;
    ed1.OnChange := CheckDirectories;
    ed1.OnExit := MoPathHelper;
    
    imgBrowse1 := TImage.Create(gb1);
    imgBrowse1.Parent := gb1;
    imgBrowse1.Picture := browse;
    imgBrowse1.Width := 18;
    imgBrowse1.Height := 18;
    imgBrowse1.ShowHint := true;
    imgBrowse1.Hint := 'Mod Organizer';
    imgBrowse1.Enabled := usingMO;
    imgBrowse1.OnClick := MoPathBrowse;
    imgBrowse1.Left := ed1.Left + ed1.Width + 8;
    imgBrowse1.Top := ed1.Top;
    
    btnFind := TButton.Create(gb1);
    btnFind.Parent := ofrm;
    btnFind.Caption := '경로 탐색';
    btnFind.ShowHint := true;
    btnFind.Hint :=
      '스크립트가 Mod Organizer의 설치 경로를 자동으로 탐색하게 됩니다. '#13
      'MO가 발견되면 경로는 자동으로 업데이트됩니다.';
    btnFind.Left := imgBrowse1.Left + imgBrowse1.Width + 24;
    btnFind.Top := lbl1.Top + btnFind.Height div 2;
    btnFind.OnClick := DetectModOrganizer;
    btnFind.Enabled := usingMO;
    
    cb2 := TCheckBox.Create(gb1);
    cb2.Parent := gb1;
    cb2.Left := 16;
    cb2.Top := lbl1.Top + lbl1.Height + 12;
    cb2.Width := 400;
    cb2.Caption := ' 일반 Asset 복사 (경로에 플러그인 이름 포함된 파일)                 ';
    cb2.Checked := copyAll;
    cb2.Enabled := usingMO;
    cb2.Hint := 
	  'Copy general assets :'#13
      '(MO 전용) MO로 병합되는 플러그인과 연결되는 Asset 파일들이 복사됩니다. '#13
      '(예: 추가 NPC의 얼굴 메쉬/화장 틴트, MCM 텍스트 파일 등) '#13
	  '플러그인과 같은 이름으로 생성되므로, 병합될 새 파일 이름을 신중하게 지어주세요.';
    cb2.ShowHint := true;
    
    rg1 := TRadioGroup.Create(ofrm);
    rg1.Parent := ofrm;
    rg1.Left := 16;
    rg1.Height := 60;
    rg1.Top := gb1.Top + gb1.Height + 12;
    rg1.Width := 560;
    rg1.Caption := '리넘버링 옵션    ';
    rg1.ClientHeight := 45;
    rg1.ClientWidth := 556;
    
    rb1 := TRadioButton.Create(rg1);
    rb1.Parent := rg1;
    rb1.Left := 26;
    rb1.Top := 18;
    rb1.Caption := 'FormID 리넘버링 안함    ';
    rb1.ShowHint := true;
    rb1.Hint := 
	  'Don''t renumber FormIDs :'#13
      '경고! 병합 전문가가 아닌 이상 이 옵션을 사용하지 마세요.'#13
      '이 옵션이 체크된 상태에서 병합하는 플러그인간에 같은 FormID가 존재하면,'#13
	  '서로 충돌하여 병합된 플러그인이 제대로 기능할 수 없습니다.';
    rb1.Width := 160;
    rb1.Checked := (rn = 0);
    
    rb2 := TRadioButton.Create(rg1);
    rb2.Parent := rg1;
    rb2.Left := rb1.Left + rb1.Width + 16;
    rb2.Top := rb1.Top;
    rb2.Caption := '(권장) 충돌 FormID 리넘버링    ';
    rb2.ShowHint := true;
    rb2.Hint :=
      'Renumber conflicting FormIDs :'#13
	  '(권장) 병합시 사용하기 최고의 옵션입니다.'#13
	  '오직 서로 충돌하는 FormID만이 리넘버링됩니다.';
    rb2.Width := 160;
    rb2.Checked := (rn = 1);
    
    rb3 := TRadioButton.Create(rg1);
    rb3.Parent := rg1;
    rb3.Left := rb2.Left + rb2.Width + 16;
    rb3.Top := rb1.Top;
    rb3.Caption := '모든 FormID 리넘버링    ';
    rb3.ShowHint := true;
    rb3.Hint :=	
      'Renumber all FormIDs :'#13
	  '충돌하는 FormID 리넘버링으로 병합이 실패할 경우 사용하세요.'#13
	  '어떠한 FormID도 충돌하는 일이 없도록 확실하게 리넘버링해 줍니다.';
    rb3.Width := 160;
    rb3.Checked := (rn = 2);
    
    rg2 := TRadioGroup.Create(ofrm);
    rg2.Parent := ofrm;
    rg2.Left := rg1.Left;
    rg2.Height := rg1.Height;
    rg2.Top := rg1.Top + rg1.Height + 16;
    rg2.Width := rg1.Width;
    rg2.Caption := '복사 옵션        ';
    rg2.ClientHeight := rg1.ClientHeight;
    rg2.ClientWidth := rg1.ClientWidth;
    
    rb4 := TRadioButton.Create(rg2);
    rb4.Parent := rg2;
    rb4.Left := 26;
    rb4.Top := 18;
    rb4.Caption := '모든 레코드 복사      ';
    rb4.ShowHint := true;
    rb4.Hint :=
      'Copy records :'#13
	  '(구버전 옵션): 이 방식은 플러그인의 모든 레코드가 복사됩니다.'#13
      'Cell처럼 다른 레코드에 걸린 레코드의 경우 두번 복사될 수 있고,'#13
	  '아주 특별한 상황에서 특정한 문제를 야기할 수 있습니다.';
    rb4.Width := 160;
    rb4.Checked := (mm = 0);
    
    rb5 := TRadioButton.Create(rg2);
    rb5.Parent := rg2;
    rb5.Left := rb4.Left + rb4.Width + 16;
    rb5.Top := rb4.Top;
    rb5.Caption := '(권장) 지능적 복사        ';
    rb5.ShowHint := true;
    rb5.Hint :=
      'Copy intelligently :'#13
	  '이 방식은 병합될 플러그인의 레코드 그룹과, 병합된 플러그인의 레코드를'#13
	  '귀납적으로 훑으며 복사합니다. 이 반복적 실행은 Cell처럼 다른 레코드에 걸린'#13
	  '레코드를 발견하면 중단되므로, 레코드가 두번 복사될 우려가 없습니다.';
    rb5.Width := 160;
    rb5.Checked := (mm = 1);
    
    rb6 := TRadioButton.Create(rg2);
    rb6.Parent := rg2;
    rb6.Left := rb5.Left + rb5.Width + 16;
    rb6.Top := rb4.Top;
    rb6.Caption := '레코드 그룹 복사      ';
    rb6.ShowHint := true;
    rb6.Hint :=
      'Copy groups :' #13
	  '(구버전 옵션) 병합될 플러그인의 레코드 그룹을 병합된 플러그인으로 복사하는'#13
      '이 방식은 레코드 복사 하나만 실패해도 그룹 전체 복사 또한 실패할 수 있고, '#13
	  '문제되는 레코드를 찾기도 어렵습니다.';
    rb6.Width := 160;
    rb6.Checked := (mm = 2);
    
    rg3 := TRadioGroup.Create(ofrm);
    rg3.Parent := ofrm;
    rg3.Left := rg1.Left;
    rg3.Height := rg1.Height;
    rg3.Top := rg2.Top + rg2.Height + 16;
    rg3.Width := rg1.Width;
    rg3.Caption := '2차 패스 복사 옵션      ';
    rg3.ClientHeight := rg1.ClientHeight;
    rg3.ClientWidth := rg1.ClientWidth;
    
    rb7 := TRadioButton.Create(rg3);
    rb7.Parent := rg3;
    rb7.Left := 26;
    rb7.Top := 18;
    rb7.Caption := '(권장) 2차 패스 없음      ';
    rb7.ShowHint := true;
    rb7.Hint :=
      'No second pass :'#13
	  '2차 패스 복사는 특정한 레코드가 제대로 기능하는 플러그인 병합 방식이었습니다.'#13
	  '최신 xEdit에서는 불필요하지만 혹시나 남겨둔 구버전 옵션입니다.';
    rb7.Width := 160;
    rb7.Checked := (sp = 0);
    
    rb8 := TRadioButton.Create(rg3);
    rb8.Parent := rg3;
    rb8.Left := rb7.Left + rb7.Width + 16;
    rb8.Top := rb7.Top;
    rb8.Caption := '처음과 같은 2차 패스      ';
    rb8.ShowHint := true;
    rb8.Hint :=
      'Second pass same as first :'#13
	  '(구버전 옵션) 이 2차 패스 방식은 위와 같은 방식으로 복사됩니다.'#13
      '하지만 2차 패스 없음 쪽을 권장합니다.';
    rb8.Width := 160;
    rb8.Checked := (sp = 1);
    
    rb9 := TRadioButton.Create(rg3);
    rb9.Parent := rg3;
    rb9.Left := rb8.Left + rb8.Width + 16;
    rb9.Top := rb7.Top;
    rb9.Caption := '그룹으로 2차 패스      ';
    rb9.ShowHint := true;
    rb9.Hint :=
      'Second pass copy by groups :'#13
	  '(구버전 옵션) 2차 패스 복사 루틴의 원래 방식입니다. 레코드 그룹 복사처럼'#13
      '이 방식도 레코드 하나가 실패하면 그룹 전체 복사가 실패하는 문제가 있습니다.';
    rb9.Width := 160;
    rb9.Checked := (sp = 2);
    
    gb2 := TGroupBox.Create(ofrm);
    gb2.Parent := ofrm;
    gb2.Left := 16;
    gb2.Height := 150;
    gb2.Top := rg3.Top + rg3.Height + 16;
    gb2.Width := 560;
    gb2.Caption := '기타 옵션      ';
    gb2.ClientHeight := 135;
    gb2.ClientWidth := 556;
    
    lbl2 := TLabel.Create(gb2);
    lbl2.Parent := gb2;
    lbl2.Left := 16;
    lbl2.Top := 25;
    lbl2.Width := 90;
    lbl2.Caption := 'Asset 저장 경로:  ';
    
    ed2 := TEdit.Create(gb2);
    ed2.Parent := gb2;
    ed2.Left := ed1.Left;
    ed2.Top := lbl2.Top;
    ed2.Width := 350;
    ed2.Caption := astPath;
    ed2.Hint := 
      '스크립트로 복사되는 Asset 파일들이 저장되는 경로입니다.'#13
	  '(추가 NPC의 얼굴 메쉬/화장, MCM 텍스트 등 플러그인 이름이 경로에 포함된 파일)'#13
	  'MO 사용자가 아닌 경우 다른 경로를 선택하세요.'#13
	  'Skyrim/Data 폴더는 권장하지 않습니다.';
    ed2.ShowHint := true;
    ed2.OnChange := CheckDirectories;
    ed2.OnExit := AssetPathHelper;
    
    imgBrowse2 := TImage.Create(gb2);
    imgBrowse2.Parent := gb2;
    imgBrowse2.Picture := browse;
    imgBrowse2.Width := 18;
    imgBrowse2.Height := 18;
    imgBrowse2.ShowHint := true;
    imgBrowse2.Hint := 'Asset 저장 경로 찾기';
    imgBrowse2.OnClick := AssetPathBrowse;
    imgBrowse2.Left := ed2.Left + ed2.Width + 8;
    imgBrowse2.Top := ed2.Top;
    
    cb3 := TCheckBox.Create(gb2);
    cb3.Parent := gb2;
    cb3.Left := lbl2.Left;
    cb3.Top := lbl2.Top +lbl2.Height + 12;
    cb3.Width := 250;
    cb3.Caption := ' 라벨 컬러링 비활성화 (재시작 필요)            ';
    cb3.ShowHint := true;
    cb3.Hint := 
      'Disable label coloring :'#13
	  '병합 메인창에서 병합할 플러그인 이름이 아무것도 보이지 않을 경우 체크하세요.'#13
	  '참고: 윈도우 개인설정에서 Aero를 활성화해도 이 문제가 교정됩니다.';
    cb3.Checked := disableColoring;
    
    cb4 := TCheckBox.Create(gb2);
    cb4.Parent := gb2;
    cb4.Left := cb3.Left;
    cb4.Top := cb3.Top + cb3.Height + 8;
    cb4.Width := 200;
    cb4.Caption := ' BSA 추출 (MO 사용시 체크해제)            ';
    cb4.ShowHint := true;
    cb4.Hint :=
      '체크하면 플러그인과 연결된 모든 BSA가 Asset 저장 경로로 추출됩니다.'#13
	  'MO 사용자라면 체크할 필요 없지만, NMM 사용자라면 체크해야 합니다.';
    cb4.Checked := extractBSAs;
    
    cb5 := TCheckBox.Create(gb2);
    cb5.Parent := gb2;
    cb5.Left := cb3.Left;
    cb5.Top := cb4.Top + cb4.Height + 8;
    cb5.Width := 200;
    cb5.Caption := ' cmd Batch로 Asset 복사            ';
    cb5.ShowHint := true;
    cb5.Hint :=
      'Batch copy assets :'#13
	  '체크하면 플러그인이 병합된 뒤 Asset가 cmd 배치로 복사됩니다.'#13
	  '커스텀 동료 모드처럼 Asset가 대용량인 모드를 병합할 경우 유용합니다.'#13
	  '모든 병합이 안전하게 처리되려면 이 옵션에 체크하세요.';
    cb5.Checked := batCopy;
    
    btnSave := TButton.Create(ofrm);
    btnSave.Parent := ofrm;
    btnSave.Caption := '저장';
    btnSave.ShowHint := true;
    btnSave.Hint :=
      '모든 병합 실행시 이 설정이 적용됩니다.';
    btnSave.ModalResult := mrOk;
    btnSave.Left := ofrm.Width div 2 - btnSave.Width - 8;
    btnSave.Top := gb2.Top + gb2.Height + 15;
    
    btnDiscard := TButton.Create(ofrm);
    btnDiscard.Parent := ofrm;
    btnDiscard.Caption := '취소';
    btnDiscard.ShowHint := true;
    btnDiscard.Hint :=
      '옵션 설정 변경이 취소됩니다.';
    btnDiscard.ModalResult := mrCancel;
    btnDiscard.Left := btnSave.Left + btnSave.Width + 16;
    btnDiscard.Top := btnSave.Top;
    
    ofrm.ActiveControl := btnSave;
    
    if ofrm.ShowModal = mrOk then begin
      if rb1.Checked then rn := 0 else
      if rb2.Checked then rn := 1 else
      if rb3.Checked then rn := 2;
      if rb4.Checked then mm := 0 else
      if rb5.Checked then mm := 1 else
      if rb6.Checked then mm := 2;
      if rb7.Checked then sp := 0 else
      if rb8.Checked then sp := 1 else
      if rb9.Checked then sp := 2;
      disableColoring := cb3.Checked;
      extractBSAs := cb4.Checked;
      batCopy := cb5.Checked;
      usingMO := cb1.Checked;
      moPath := ed1.Caption;
      copyAll := cb2.Checked;
      astPath := ed2.Caption;
      SaveSettings;
    end;
  finally
    ofrm.Free;
  end;
end;

//=========================================================================
// AssetHelperBrowse: Show user their asset destination directory
procedure afrm.AssetHelperBrowse;
begin
  ShellExecute(TForm(frmMain).Handle, 'open', astPath, '', '', SW_SHOWNORMAL)
end;

//=========================================================================
// AssetHelper: Helps the user set up their asset destination folder
function AssetHelper: boolean;
var
  afrm: TForm;
  lbl1: TLabel;
  btnBrowse, btnOk, btnCancel: TButton;
begin
  Result := true;
  if (astPath = DataPath) then begin
    if firstrun then 
      Result := (MessageDlg(
        '스크립트로 복사되는 Asset 저장 경로를 Skyrim/Data 폴더로 설정하면, '+
        '점차 Data 폴더 안의 파일이 엉망진창이 되므로 권장하지 않습니다. '+
        '정말 계속하시겠습니까? ',  mtCustom, [mbYes,mbNo], 0) = mrYes);
  end
  else if not IsDirectoryEmpty(astPath) then begin
    afrm := TForm.Create(nil);
    try
      afrm.Caption := 'Asset 저장 경로 도우미';
      afrm.Width := 300;
      afrm.Position := poScreenCenter;
      afrm.Height := 130;
      
      lbl1 := TLabel.Create(afrm);
      lbl1.Parent := afrm;
      lbl1.Top := 8;
      lbl1.Width := 250;
      lbl1.Height := 30;
      lbl1.AutoSize := False;
      lbl1.Wordwrap := True;
      lbl1.Caption := '저장 경로 안에 파일이 있습니다. 병합 전에 비워두실 것을 권장합니다.';
      lbl1.Left := 8;
      
      btnOk := TButton.Create(afrm);
      btnOk.Parent := afrm;
      btnOk.Caption := '병합 실행';
      btnOk.ShowHint := true;
      btnOk.Hint := '병합을 실행합니다.';
      btnOk.ModalResult := mrOk;
      btnOk.Left := (afrm.Width - 10) div 2 - (btnOk.Width * 1.5 + 8);
      btnOk.Top := lbl1.Top + lbl1.Height + 16;
      
      btnBrowse := TButton.Create(afrm);
      btnBrowse.Parent := afrm;
      btnBrowse.Caption := '폴더 열기';
      btnBrowse.ShowHint := true;
      btnBrowse.Hint := 'Asset 저장 경로를 윈도우 탐색기로 열어봅니다.';
      btnBrowse.OnClick := AssetHelperBrowse;
      btnBrowse.Left := btnOk.Left + btnOk.Width + 8;
      btnBrowse.Top := btnOk.Top;
      
      btnCancel := TButton.Create(afrm);
      btnCancel.Parent := afrm;
      btnCancel.Caption := '취소';
      btnCancel.ShowHint := true;
      btnCancel.Hint := '병합을 취소합니다.';
      btnCancel.ModalResult := mrCancel;
      btnCancel.Left := btnBrowse.Left + btnBrowse.Width + 8;
      btnCancel.Top := btnOk.Top;
      
      if afrm.ShowModal = mrCancel then
        Result := false;
    finally
      afrm.Free;
    end;
  end;
end;

//=========================================================================
// PluginForm: Provides user with information on a specific plugin
procedure PluginForm(Sender: TObject);
var
  pfrm: TForm;
  fn: string;
  lbl: TLabel;
  sb: TScrollBox;
  pnl: TPanel;
  btnOk: TButton;
  f: IInterface;
  slDefinitions: TStringList;
  definitions: string;
  i, j: integer;
begin
  fn := TLabel(Sender).Caption;
  fn := Copy(fn, Pos(']', fn) + 3, Length(fn));
  
  f := FileByName(fn);
  if not Assigned(f) then exit;
  pfrm := TForm.Create(nil);
  try
    pfrm.Caption := fn+' 정보';
    pfrm.Width := 425;
    pfrm.Height := 450;
    pfrm.Position := poScreenCenter;
    
    // construct labels
    lbl := ConstructLabel(pfrm, pfrm, 8, 8, 0, 300, '파일 이름: '+fn);
    lbl := ConstructLabel(pfrm, pfrm, lbl.Top + lbl.Height + 8, lbl.Left, 0, 300, 
      '플러그인 내 레코드: '+IntToStr(RecordCount(f)));
    lbl := ConstructLabel(pfrm, pfrm, lbl.Top + lbl.Height + 8, lbl.Left, 0, 300, 
      '병합 스크립트 버전: '+vs);
    lbl.ShowHint := true;
    lbl.Hint := '사용중인 병합 플러그인의 버전입니다.';
    lbl := ConstructLabel(pfrm, pfrm, lbl.Top + lbl.Height + 8, lbl.Left, 0, 300, 
      '리포트: ');
    
    sb := TScrollBox.Create(pfrm);
    sb.Parent := pfrm;
    sb.Top := lbl.Top + lbl.Height + 16;
    sb.Align := alBottom;
    sb.Height := 320;
    
    slDefinitions := TStringList.Create;
    slDefinitions.StrictDelimiter := true;
    slDefinitions.Delimiter := ';';
    definitions := GetAllDefinitions(f);
    slDefinitions.DelimitedText := definitions;
    
    // construct report labels
    lbl := TLabel.Create(pfrm);
    lbl.Top := 0;
    lbl.Height := 0;
    i := 0;
    while i + 6 <= slDefinitions.Count - 1 do begin
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 12, 8, 0, 360, 
        '파일 이름:   '+slDefinitions[i]);
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 4, 8, 0, 360, 
        '레코드 개수:   '+slDefinitions[i+1]);
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 4, 8, 0, 360, 
        '병합 스크립트 버전:   '+slDefinitions[i+2]);
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 4, 8, 0, 360, 
        '평균 평가:   '+slDefinitions[i+3]);
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 4, 8, 0, 360, 
        '리포트 개수:   '+slDefinitions[i+4]);
      lbl := ConstructLabel(sb, sb, lbl.Top + lbl.Height + 4, 8, 0, 360, 
        '노트:'#13'    '+ StringReplace(slDefinitions[i+5], '@13', #13'    ', [rfReplaceall]));
      i := i + 6;
    end;
    
    pfrm.ShowModal;
  finally
    pfrm.Free;
  end;
end;

//=========================================================================
// MergeForm: Provides user with options for merging
procedure MergeForm;
var
  mfrm: TForm;
  btnOk, btnCancel, btnFocus: TButton;
  imgOptions: TImage;
  lbl1, lbl2, lbl3: TLabel;
  pnl: TPanel;
  sb: TScrollBox;
  i, j, k, height, m: integer;
  holder: TObject;
  masters, e, f: IInterface;
  s, definition, hint: string;
  slDefinition: TStringList;
begin
  LoadSettings;
  mfrm := TForm.Create(nil);
  try
    mfrm.Caption := '플러그인 병합';
    mfrm.Width := 425;
    mfrm.Position := poScreenCenter;
    for i := 0 to FileCount - 1 do begin
      s := GetFileName(FileByLoadOrder(i));
      if Pos(s, bethesdaFiles) > 0 then Continue;
      Inc(m);
    end;
    height := m*25 + 120;
    if height > (Screen.Height - 100) then begin
      mfrm.Height := Screen.Height - 100;
      sb := TScrollBox.Create(mfrm);
      sb.Parent := mfrm;
      sb.Height := Screen.Height - 210;
      sb.Align := alTop;
      holder := sb;
    end
    else begin
      mfrm.Height := height;
      holder := mfrm;
    end;

    lbl1 := TLabel.Create(holder);
    lbl1.Parent := holder;
    lbl1.Top := 8;
    lbl1.Left := 8;
    lbl1.AutoSize := False;
    lbl1.Wordwrap := True;
    lbl1.Width := 500;
    lbl1.Height := 50;
    lbl1.Caption := 
	'병합할 플러그인들을 선택하세요. (나사 아이콘: 고급 옵션 / 이름 클릭시: 리포트 보기)         ';
    
    // create file list
    for i := 0 to FileCount - 1 do begin
      s := GetFileName(FileByIndex(i));
      if (Pos(s, bethesdaFiles) > 0) or (s = '') then Continue;
      j := 25 * k;
      Inc(k);
      
      // load definition
      slDefinition := TStringList.Create;
      slDefinition.StrictDelimiter := true;
      slDefinition.Delimiter := ';';
      definition := GetDefinition(FileByIndex(i), true, true);
      if (definition = '') then definition := GetDefinition(FileByIndex(i), true, false);
      if (definition = '') then definition := GetDefinition(FileByIndex(i), false, false);
      slDefinition.DelimitedText := definition;
      hint := GetDefinitionHint(slDefinition);
      
      // set up checkbox
      cbArray[i] := TCheckBox.Create(holder);
      cbArray[i].Parent := holder;
      cbArray[i].Left := 24;
      cbArray[i].Top := 40 + j;
      cbArray[i].Width := 40;
      cbArray[i].ShowHint := true;
      cbArray[i].Hint := hint;
        
      if (slSelectedFiles.IndexOf(s) > - 1) then 
        cbArray[i].Checked := True;
      
      // set up label
      lbArray[i] := TLabel.Create(holder);
      lbArray[i].Parent := holder;
      lbArray[i].Left := 44;
      lbArray[i].Top := cbArray[i].Top;
      lbArray[i].Caption := '  [' + IntToHex64(i + 1, 2) + ']  ' + s;
      lbArray[i].OnClick := PluginForm; 
      lbArray[i].ShowHint := true;
      lbArray[i].Hint := hint;
      if not disableColoring then begin
        lbArray[i].Font.Color := GetMergeColor(slDefinition);
        if slDefinition.Count > 5 then
          lbArray[i].Font.Style := lbArray[i].Font.Style + [fsbold];
      end;
      
      // free definition
      slDefinition.Free;
    end;
    
    if holder = sb then begin
      lbl2 := TLabel.Create(holder);
      lbl2.Parent := holder;
      lbl2.Top := j + 60;
    end;
    
    pnl := TPanel.Create(mfrm);
    pnl.Parent := mfrm;
    pnl.BevelOuter := bvNone;
    pnl.Align := alBottom;
    pnl.Height := 50;
    
    imgOptions := TImage.Create(pnl);
    imgOptions.Parent := pnl;
    imgOptions.Picture := gear;
    imgOptions.Width := 24;
    imgOptions.Height := 24;
    imgOptions.ShowHint := true;
    imgOptions.Hint := '고급 옵션';
    imgOptions.OnClick := AdvancedOptions;
    imgOptions.Left := mfrm.Width - 50;
    imgOptions.Top := pnl.Height - 40;
    
    btnOk := TButton.Create(mfrm);
    btnOk.Parent := pnl;
    btnOk.Caption := '확인';
    btnOk.ModalResult := mrOk;
    btnOk.Left := 120;
    btnOk.Top := pnl.Height - 40;
    
    btnCancel := TButton.Create(mfrm);
    btnCancel.Parent := pnl;
    btnCancel.Caption := '취소';
    btnCancel.ModalResult := mrCancel;
    btnCancel.Left := btnOk.Left + btnOk.Width + 16;
    btnCancel.Top := btnOk.Top;
    
    mfrm.ActiveControl := btnOk;
    
    if (firstRun) then AdvancedOptions;
    
    if mfrm.ShowModal = mrOk then begin
      for i := 0 to FileCount - 1 do begin
        f := FileByIndex(i);
        s := GetFileName(f);
        if Pos(s, bethesdaFiles) > 0 then Continue;        
        
        if cbArray[i].State = cbChecked then begin
          slMerge.AddObject(s, TObject(GetLoadOrder(f)));
          slMasters.Add(s);
          // add masters from files to be merged
          masters := ElementByName(ElementByIndex(f, 0), 'Master Files');
          for j := 0 to ElementCount(masters) - 1 do begin
            e := ElementByIndex(masters, j);
            s := GetElementNativeValues(e, 'MAST');
            slMasters.Add(s);
          end;
        end;
      end;
    end;
  finally
    mfrm.Free;
  end;
end;

//======================================================================
// ShowDetails: Enables the visibilty of the TMemo log
procedure ShowDetails;
begin
  frm.Height := 600;
  frm.Position := poScreenCenter;
  memo.Height := frm.Height - 150;
  btnDetails.Visible := false;
  memo.Visible := true;
end;


{*************************************************************************}
{**************************** Merge Functions ****************************}
{*************************************************************************}

//=========================================================================
// AddCopyOperation: Adds a copy operation to batch
procedure AddCopyOperation(src, dst: string);
begin
  batch.Add('copy /Y "'+src+'" "'+dst+'"');
end;

//=========================================================================
// FindFolder: looks for a folder matching the given name at the given path
function FindFolder(path: string; filename: string): string;
var
  info: TSearchRec;
begin
  Result := '';
  if FindFirst(path+'*', faDirectory, info) = 0 then begin
    repeat
      if Lowercase(info.Name) = Lowercase(filename) then begin
        Result := path + info.Name +'\';
        break;
      end;
    until FindNext(info) <> 0;
  end;
end;

//=========================================================================
// CopyGeneralAssets: copies all assets from a matching Mod Organizer folder
procedure CopyGeneralAssets(filename: string);
var
  ignore: TStringList;
  rec: TSearchRec;
  src, dst, modPath, exclusions: string;
begin
  // construct ignore stringlist
  ignore := TStringList.Create;
  if not (batCopy and useRobocopy) then begin
    ignore.Add('facegendata');
    ignore.Add('voice');
    ignore.Add('translations');
  end;
  ignore.Add('meta.ini');
  ignore.Add('*.esp');
  ignore.Add('*.esm');
  if extractBSAs then begin
    ignore.Add('*.bsa');
    ignore.Add('*.bsl');
  end;
  
  // find mod directory in Mod Organizer's mods folder
  if debugAssetCopying then LogMessage('    파일 탐색 중: '+filename+' 경로: '+moPath);
  if FindFirst(moPath + 'mods\*', faDirectory, rec) = 0 then begin
    repeat
      if (Pos('.', rec.Name) = 1) then Continue;
      if debugAssetCopying then LogMessage('    ...탐색 중: '+moPath+'mods\'+rec.Name);
      if (FileExists(moPath + 'mods\' + rec.Name + '\' + filename)) then begin
        modPath := moPath + 'mods\'+ rec.Name;
        break;
      end;
    until FindNext(rec) <> 0;
    
    FindClose(rec);
  end;
  
  // copy assets from folder
  if (modPath <> '') and (slCopiedFrom.IndexOf(modPath) = -1) then begin
    slCopiedFrom.Add(modPath);
    LogMessage('    다음 폴더에서 Asset 복사 중 "'+modPath+'"');
    if batCopy and useRobocopy then 
      batch.Add('>> ..\logs\merge_'+fdt+'.txt robocopy "'+modPath+'" "'+RemoveFromEnd(astPath, '\')+'" /e /xf '+
      StringReplace(ignore.Text, #13#10, ' ', [rfReplaceAll])+' /xd facegendata voice translations')
    else if batCopy then 
      BatchCopyDirectory(modPath, astPath, ignore, batch, debug)
    else 
      CopyDirectory(modPath, astPath, ignore, debug);
  end;
  
  // free ignore stringlist
  ignore.Free;
end;

//=========================================================================
// CopyAssets: copies assets in filename specific directories 
procedure CopyActorAssets(path: string; mergeIndex: integer);
var
  info: TSearchRec;
  src, dst, newForm, oldForm, srcPath, dstPath: string;
  index: integer;
begin
  // see if there is a folder matching the filename
  srcPath := FindFolder(path, slMerge[mergeIndex]);
  
  // if no folder found, exit
  if (srcPath = '') then exit;
  
  // prepare destination
  dstPath := StringReplace(path + GetFileName(mgf) + '\', DataPath, astPath, [rfReplaceAll]);
  dstPath := StringReplace(dstPath, temp, astPath, [rfReplaceAll]);
  ForceDirectories(dstPath);
  
  // copy all assets
  index := -1;
  LogMessage('        다음 폴더에서 Asset 복사 중: "'+srcPath+'"');
  LogMessage('        다음 폴더로 Asset 복사 중: "'+dstPath+'"');
  if FindFirst(srcPath+'*', faAnyFile, info) = 0 then begin
    repeat
      // skip . and ..
      if (Length(info.Name) >= 8) then begin
        src := info.Name;
        oldForm := Copy(src, 1, 8);
        if debugAssetCopying then LogMessage('          복사 시도 중: '+src+', oldForm = '+oldForm);
        if (rn > 0) then
          index := TStringList(OldForms[mergeIndex]).IndexOf(oldForm);
        // asset not renumbered
        if (index = -1) then begin
          if debugAssetCopying then LogMessage('          Asset 리넘버링 안됨.');
          dst := info.Name;
          LogMessage('            Asset 복사 중: "'+src+'"');
          if batCopy then AddCopyOperation(srcPath+src, dstPath+dst)
          else CopyFile(PChar(srcPath+src), PChar(dstPath+dst), true);
          //ResourceCopy('Data', srcPath + src, dstPath + dst);
          //wCopyFile(srcPath + src, dstPath + dst, true);
        end
        // asset renumbered
        else begin
          if debugAssetCopying then LogMessage('          Asset 리넘버링 완료: '+TStringList(NewForms[mergeIndex]).Strings[index]);
          newForm := '00' + Copy(TStringList(NewForms[mergeIndex]).Strings[index], 3, 6);
          dst := StringReplace(info.Name, oldForm, newForm, [rfReplaceAll]);
          if batCopy then AddCopyOperation(srcPath+src, dstPath+dst)
          else CopyFile(PChar(srcPath+src), PChar(dstPath+dst), true);
          //ResourceCopy('Data', srcPath + src, dstPath + dst);
          //wCopyFile(srcPath + src, dstPath + dst, true);
          LogMessage('            Asset 복사 중: "'+src+'" 에서 "'+dst+'"');
        end;
      end;
    until FindNext(info) <> 0;
    FindClose(info);
  end;
end;

//=========================================================================
// CopyVoiceAssets: copies voice assets in filename specific directories
procedure CopyVoiceAssets(path: string; mergeIndex: integer);
var
  info, info2: TSearchRec;
  src, dst, newForm, srcPath, dstPath, oldForm: string;
  index: integer;
begin
  // see if there is a folder matching the filename
  srcPath := FindFolder(path, slMerge[mergeIndex]);
  
  // if no folder found, exit
  if (srcPath = '') then exit;
  
  // prepare destination
  dstPath := StringReplace(path + GetFileName(mgf) + '\', DataPath, astPath, [rfReplaceAll]);
  dstPath := StringReplace(dstPath, temp, astPath, [rfReplaceAll]);
  ForceDirectories(dstPath);
  
  // copy subfolders and their contents
  index := -1;
  LogMessage('        다음 폴더에서 Voice Asset 복사 중: "'+srcPath+'"');
  LogMessage('        다음 폴더로 Voice Asset 복사 중: "'+dstPath+'"');
  if FindFirst(srcPath+'*', faDirectory, info) = 0 then begin
    repeat
      if (Pos('.', info.Name) <> 1) then begin
        CreateDir(dstPath+info.Name);
        // copy contents of subdirectory into new directory
        if FindFirst(srcPath+info.Name+'\'+'*', faAnyFile, info2) = 0 then begin
          repeat
            // skip '.' and '..'
            if (Length(info2.Name) >= 8) then begin
              src := info.Name + '\' + info2.Name;
              oldForm := Copy(info2.Name, Pos('.', info2.Name) - 10, 8);
              if (rn > 0) then 
                index := TStringList(OldForms[mergeIndex]).IndexOf(oldForm);
              // asset not renumbered
              if (index = -1) then begin
                dst := info.Name + '\' + info2.Name;
                LogMessage('            Asset 복사 중: "'+src+'"');
                if batCopy then AddCopyOperation(srcPath+src, dstPath+dst)
                else CopyFile(PChar(srcPath+src), PChar(dstPath+dst), true);
                //ResourceCopy('Data', srcPath + src, dstPath + dst);
                //wCopyFile(srcPath + src, dstPath + dst, true);
              end
              // asset renumbered
              else begin
                newForm := '00' + Copy(TStringList(NewForms[mergeIndex]).Strings[index], 3, 6);
                dst := StringReplace(src, oldForm, newForm, [rfReplaceAll]);
                if batCopy then AddCopyOperation(srcPath+src, dstPath+dst)
                else CopyFile(PChar(srcPath+src), PChar(dstPath+dst), true);
                //ResourceCopy('Data', srcPath + src, dstPath + dst);
                //wCopyFile(srcPath + src, dstPath + dst, true);
                LogMessage('            Asset 복사 중: "'+src+'" 에서 "'+dst+'"');
              end;
            end;
          until FindNext(info2) <> 0;
        end;
      end;
    until FindNext(info) <> 0;
  end;
end;

//=========================================================================
// CopyTranslations: copies MCM translation files
procedure CopyTranslations(path: string; mergeIndex: integer);
var
  info: TSearchRec;
  t, fn, srcPath, dstPath: string;
  slSrc: TStringList;
  index: integer;
begin
  fn := slMerge[mergeIndex];
  fn := Lowercase(Copy(fn, 1, Length(fn) - 4)); // trim .esp off
  if debugMCM then LogMessage('     연결된 MCM 텍스트 복사 중: '+fn);
  
  // search for translation files
  if FindFirst(path+'*.txt', faAnyFile and faDirectory, info) = 0 then begin
    repeat
      // translation file found
      if (Pos(fn, Lowercase(info.Name)) = 1) then begin
        t := StringReplace(Lowercase(info.Name), fn, '', [rfReplaceAll]);
        index := slTranslations.IndexOf(t);
        // other translation files for same language found, concatenate
        if index > -1 then begin
          slSrc := TStringList.Create;
          if debugMCM then LogMessage('            LoadFromFile: "'+path+info.Name+'"');
          slSrc.LoadFromFile(path+info.Name);
          slArray[index].Text := slArray[index].Text + #13#10#13#10 + slSrc.Text;
          slSrc.Free;
        end
        // add new translation to stringlist
        else begin
          slArray[slTranslations.Count] := TStringList.Create;
          if debugMCM then LogMessage('            LoadFromFile: "'+path+info.Name+'"');
          slArray[slTranslations.Count].LoadFromFile(path+info.Name);
          slTranslations.Add(t);
        end;
        LogMessage('          MCM 텍스트 복사 중: "'+info.Name+'"');
      end;
    until FindNext(info) <> 0;
  end;
end;

//=========================================================================
// SaveTranslations
procedure SaveTranslations(path: string);
var
  i: integer;
  output: string;
begin
  // terminate if we have no translation files to save
  if slTranslations.Count = 0 then
    exit;
  
  // use MO's overwrite folder as destination if user is using MO
  path := StringReplace(path, DataPath, astPath, [rfReplaceAll]);
  ForceDirectories(path);
  
  // save all new translation files
  for i := 0 to slTranslations.Count - 1 do begin
    output := path+Copy(GetFileName(mgf), 1, Length(GetFileName(mgf)) - 4) + slTranslations[i];
    if debugMCM then 
      LogMessage('            MCM 텍스트 내보내기: "'+output);
    slArray[i].SaveToFile(output);
    slArray[i].Free;
  end;
end;

//=========================================================================
// CopyFileSpecificAssets: performs all file specific asset copying
procedure CopyFileSpecificAssets(i: integer);
begin
  // extract assets from BSAs if necessary
  if (Pos('.esp', slMerge[i]) > 0) then
    bsaName := StringReplace(slMerge[i], '.esp', '.bsa', [rfReplaceAll]);
  if (Pos('.esm', slMerge[i]) > 0) then
    bsaName := StringReplace(slMerge[i], '.esm', '.bsa', [rfReplaceAll]);
  if (Pos('.bsa', bsaName) > 0) and FileExists(DataPath + bsaName) then begin
    ExtractPathBSA(DataPath + bsaName, temp, 'textures\actors\character\facegendata\facetint\');
    ExtractPathBSA(DataPath + bsaName, temp, 'meshes\actors\character\facegendata\facegeom\');
    ExtractPathBSA(DataPath + bsaName, temp, 'sound\voice');
    ExtractPathBSA(DataPath + bsaName, temp, 'interface\translations\');
  end;
  
  // copy loose File/FormID specific assets
  CopyActorAssets(DataPath + 'Textures\Actors\Character\FacegenData\facetint\', i); // copy actor textures
  CopyActorAssets(DataPath + 'Meshes\actors\character\facegendata\facegeom\', i); // copy actor meshes
  CopyVoiceAssets(DataPath + 'Sound\Voice\', i); // copy voice assets
  CopyTranslations(DataPath + 'Interface\Translations\', i); // copy MCM translation files
  
  // copy archived File/FormID specific assets
  CopyActorAssets(temp + 'Textures\Actors\Character\FacegenData\facetint\', i); // copy actor textures
  CopyActorAssets(temp + 'Meshes\actors\character\facegendata\facegeom\', i); // copy actor meshes
  CopyVoiceAssets(temp + 'Sound\Voice\', i); // copy voice assets
  CopyTranslations(temp + 'Interface\Translations\', i); // copy MCM translation files
end;

//=========================================================================
// CopyElement: copies an element to the merged file
function CopyElement(e: IInterface; remove: boolean): boolean;
var
  cr: IInterface;
begin
  Result := false;
  // records that fulfill the conditions below throw a duplicate formID
  // error when copied
  if (geev(e, pFlag) = '1') and (geev(WinningOverride(e), pFlag) = '') then begin
    //AddMessage('건너뜀: '+Name(e));
    exit;
  end;
  
  // correct Tamriel camera data
  if (geev(e, 'EDID') = 'Tamriel') then begin
    Remove(ElementByPath(e, 'MNAM'));
    Add(e, 'MNAM', True);
    seev(e, 'MNAM\Cell Coordinates\NW Cell\X', '-30');
    seev(e, 'MNAM\Cell Coordinates\NW Cell\Y', '15');
    seev(e, 'MNAM\Cell Coordinates\SE Cell\X', '40');
    seev(e, 'MNAM\Cell Coordinates\SE Cell\Y', '-40');
    seev(e, 'MNAM\Cell Coordinates\SE Cell\Y', '-40');
    seev(e, 'MNAM\Camera Data\Min Height', '50000');
    seev(e, 'MNAM\Camera Data\Max Height', '80000');
    seev(e, 'MNAM\Camera Data\Initial Pitch', '50');
  end;
  
  // attempt to copy record to merged file, alert user on exception
  try
    cr := wbCopyElementToFile(e, mgf, False, True);
    if debug then LogMessage('        복사 중: '+SmallName(e));
    Inc(rCount);
    if remove then RemoveNode(cr);
    Result := true;
  except
    on x : Exception do begin
      Result := false;
      if (Pos('Duplicate FormID', x.Message) = 1) then begin
        if (Signature(e) <> 'GRUP') then 
          LogMessage('        복사 실패: '+Name(e)+'; '+x.Message);
      end
      else begin
        LogMessage('        복사 실패: '+Name(e)+'; '+x.Message);
        slFails.Add(FullPath(e)+'; '+x.Message);
      end;
    end;
  end;
end;

//=========================================================================
// MergeByRecords: merges by copying records
procedure MergeByRecords(g: IInterface; remove: boolean);
var
  i: integer;
  e: IInterface;
begin
  for i := 0 to RecordCount(g) - 1 do begin
    e := RecordByIndex(g, i);
    if Signature(e) = 'TES4' then Continue;
    CopyElement(e, remove);
  end;
end;

//=========================================================================
// MergeIntelligently: merges by not copying group records
procedure MergeIntelligently(g: IInterface; remove: boolean);
var
  i: integer;
  e: IInterface;
begin
  for i := 0 to ElementCount(g) - 1 do begin
    e := ElementByIndex(g, i);
    if Signature(e) = 'TES4' then Continue;
    if Signature(e) = 'GRUP' then begin
      if (Pos('GRUP Cell', Name(e)) = 1) or (Pos('GRUP Exterior Cell', Name(e)) = 1) then begin
        if not CopyElement(e, remove) then
        MergeIntelligently(e, remove);
      end
      else MergeIntelligently(e, remove);
    end
    else CopyElement(e, remove);
  end;
end;

//=========================================================================
// MergeByGroups: merges by copying entire group records
procedure MergeByGroups(g: IInterface; remove: boolean);
var
  i: integer;
  e: IInterface;
begin
  for i := 0 to ElementCount(g) - 1 do begin
    e := ElementByIndex(g, i);
    if Signature(e) = 'TES4' then Continue;
    CopyElement(e, remove);
  end;
end;

// ========================================================================
// FindHighestFormID: find highest formID in files to be merged
function FindHighestFormID(): Cardinal;
var
  i, j: integer;
  x: Cardinal;
  e, f: IInterface;
begin
  Result := 0;
  
  // check files to be merged for highest formID
  for i := 0 to slMerge.Count - 1 do begin
    f := FileByLoadOrder(Integer(slMerge.Objects[i]));
    for j := 0 to RecordCount(f) - 1 do begin
      e := RecordByIndex(f, j);
      // exclude override records, include injected records
      if not Equals(e, MasterOrSelf(e)) then Continue;
      x := FileFormID(e);
      if x > Result then Result := x;
    end;
  end;
  
  // check merge file for a higher form ID
  for i := 0 to RecordCount(mgf) - 1 do begin
    e := RecordByIndex(mgf, i);
    // exclude override records, include injected records
    if not Equals(e, MasterOrSelf(e)) then Continue;
    x := FileFormID(e);
    if x > Result then Result := x;
  end;
end;

//=========================================================================
// RenumberRecord: renumbers a record
procedure RenumberRecord(e: IInterface; NewFormID: Cardinal);
var
  OldFormID, prc: Cardinal;
begin
  OldFormID := GetLoadOrderFormID(e);
  // change references, then change form
  prc := 0;
  while ReferencedByCount(e) > 0 do begin
    if prc = ReferencedByCount(e) then break;
    prc := ReferencedByCount(e);
    CompareExchangeFormID(ReferencedByIndex(e, 0), OldFormID, NewFormID);
  end;
  SetLoadOrderFormID(e, NewFormID);
end;

//=========================================================================
// RenumberConflicting: renumber only conflicting FormIDs.
procedure RenumberConflicting(pb: TProgressBar);
var
  i, j, k, rc, pre, ndx: integer;
  HighestFormID, NewFormID, BaseFormID, InjectFormID, offset: Cardinal;
  e, f: IInterface;
  self: boolean;
  loadFormID, fileFormID: String;
  slAllForms: TStringList;
begin
  pb.Position := 1;
  LogMessage(#13#10+'병합 전 충돌 FormID를 리넘버링하는 중...');
  
  // add formIDs from merge file to UsedFormIDs
  for i := 0 to RecordCount(mgf) - 1 do begin
    e := RecordByIndex(mgf, i);
    if Signature(e) = 'TES4' then Continue;
    if IsOverrideRecord(e) then Continue;
    fileFormID := '00' + Copy(HexFormID(e), 3, 6);
    ndx := StrToInt('$' + fileFormID);
    UsedFormIDs[ndx] := 1;
  end;
  
  // find a safe NewFormID to start at
  HighestFormID := FindHighestFormID();
  BaseFormID := HighestFormID + 1024;
  
  // form id renumbering for each file
  for i := 0 to slMerge.Count - 1 do begin
    f := FileByLoadOrder(Integer(slMerge.Objects[i]));
    RC := RecordCount(f) - 1;
    LogMessage('    다음 파일의 레코드 리넘버링 중: '+GetFileName(f));
    OldForms.Add(TStringList.Create);
    NewForms.Add(TStringList.Create);
    
    // create records array for file because the indexed order of records changes 
    // as we alter their formIDs
    for j := 0 to RC do
      Records[j] := RecordByIndex(f, j);
    
    // set newformID to use the load order of the file currently being processed.
    offset := Integer(slMerge.Objects[i]) * $01000000;
    NewFormID := BaseFormID + offset;

    // renumber the records in the file
    for j := 0 to RC do begin
      e := Records[j];
      // skip header record
      if (Signature(e) = 'TES4') then Continue;
      
      // prepare variables
      loadFormID := HexFormID(e);
      fileFormID := '00' + Copy(loadFormID, 3, 6);
      ndx := StrToInt('$' + fileFormID);
      
      // skip override records 
      if IsOverrideRecord(e) then begin
        UsedFormIDs[ndx] := 1;
        TStringList(OldForms[i]).Add(loadFormID);
        TStringList(NewForms[i]).Add(loadFormID);
        Continue;
      end;
      
      // if not conflicting FormID, add to list and continue.
      if (UsedFormIDs[ndx] = 0) then begin
        UsedFormIDs[ndx] := 1;
        TStringList(OldForms[i]).Add(loadFormID);
        TStringList(NewForms[i]).Add(loadFormID);
        Continue;
      end
      // special renumbering for injected formIDs
      else if (IsInjected(e)) then begin
        InjectFormID := StrToInt('$' + Copy(HexFormID(e), 1, 2)) * $01000000 + BaseFormID;
        //if debugRenumbering then 
          LogMessage(Format('        Injected FormID 변경 중: [%s] , %s', 
            [IntToHex64(InjectFormID, 8), SmallName(e)]));
        RenumberRecord(e, InjectFormID);
        TStringList(OldForms[i]).Add(fileFormID);
        TStringList(NewForms[i]).Add(IntToHex64(InjectFormID, 8));
        
        Inc(rCount);
        // increment formid
        Inc(BaseFormID);
        Inc(NewFormID);
      end
      // else renumber it
      else begin
        if debugRenumbering then 
          LogMessage(Format('        FormID 변경 중: [%s], %s', 
            [IntToHex64(NewFormID, 8), SmallName(e)]));
        RenumberRecord(e, NewFormID);
        TStringList(OldForms[i]).Add(fileFormID);
        TStringList(NewForms[i]).Add(IntToHex64(NewFormID, 8));
        
        Inc(rCount);
        // increment formid
        Inc(BaseFormID);
        Inc(NewFormID);
      end;
    end;
    
    // copy file specific assets
    CopyFileSpecificAssets(i);
    
    pb.Position := pb.Position + 29/slMerge.Count;
  end;
  NextObjectID := BaseFormID + 1;
  SaveTranslations(DataPath + 'Interface\Translations\');
end;

//=========================================================================
// RenumberNew: the new renumbering method, for 3.033 and above
procedure RenumberAll(pb: TProgressBar);
var
  i, j, k, rc, pre: integer;
  HighestFormID, OldFormID, NewFormID, BaseFormID, offset, x, prc: Cardinal;
  e, f: IInterface;
  self: boolean;
  fileFormID, loadFormID: String;
begin
  pb.Position := 1;
  LogMessage(#13#10+'병합 전 FormID 리넘버링 중...');
  
  // find a safe NewFormID to start at
  HighestFormID := FindHighestFormID();
  BaseFormID := HighestFormID + 1024;
  
  // form id renumbering for each file
  for i := 0 to slMerge.Count - 1 do begin
    f := FileByLoadOrder(Integer(slMerge.Objects[i]));
    RC := RecordCount(f) - 1;
    LogMessage('    다음 파일의 레코드 리넘버링 중: '+GetFileName(f));
    OldForms.Add(TStringList.Create);
    NewForms.Add(TStringList.Create);
    
    // create records array for file because the indexed order of records changes 
    // as we alter their formIDs
    for j := 0 to RC do
      Records[j] := RecordByIndex(f, j);
      
    // set newformID to use the load order of the file currently being processed.
    offset := Integer(slMerge.Objects[i]) * $01000000;
    NewFormID := BaseFormID + offset;
    
    // renumber the records in the file
    for j := 0 to RC do begin
      e := Records[j];
      if (Signature(e) = 'TES4') then Continue;
      
      // prepare variables
      loadFormID := HexFormID(e);
      fileFormID := '00' + Copy(loadFormID, 3, 6);
      
      // skip non-file records (override and injected)
      if not (IsLocalRecord(e)) then begin
        TStringList(OldForms[i]).Add(loadFormID);
        TStringList(NewForms[i]).Add(loadFormID);
        Continue;
      end;
      
      // print log message first, then change references, then change form
      OldFormID := GetLoadOrderFormID(e);
      if debugRenumbering then 
        LogMessage(Format('        FormID 변경 중: [%s], %s', 
        [IntToHex64(NewFormID, 8), SmallName(e)]));
      RenumberRecord(e, NewFormID);
      TStringList(OldForms[i]).Add(fileFormID);
      TStringList(NewForms[i]).Add(IntToHex64(NewFormID, 8));
      
      Inc(rCount);
      // increment formid
      Inc(BaseFormID);
      Inc(NewFormID);
    end;
    
    // copy file specific assets
    CopyFileSpecificAssets(i);
    
    pb.Position := pb.Position + 29/slMerge.Count;
  end;
  NextObjectID := BaseFormID + 1;
  SaveTranslations(DataPath + 'Interface\Translations\');
end;


//=========================================================================
// FreeMemory: frees memory used by script
procedure FreeMemory;
begin
  slMerge.Free; slMasters.Free; slSelectedFiles.Free; slFails.Free; 
  slTranslations.Free; slDictionary.Free; slMgfMasters.Free; 
  slCopiedFrom.Free; batch.Free; NewForms.Free; OldForms.Free;
  gear.Free; browse.Free;
end;

{**************************************************************************}
{**************************** Script Execution ****************************}
{**************************************************************************}

//=========================================================================
// Initialize
function Initialize: integer;
begin
  // welcome messages
  AddMessage(#13#10#13#10);
  AddMessage(dashes);
  AddMessage('병합 플러그인 '+vs+': 모드 플러그인들을 병합해 줍니다.');
  AddMessage(dashes);
 
  // stringlist creation
  slSelectedFiles := TStringList.Create;
  slMerge := TStringList.Create;
  slFails := TStringList.Create;
  slMasters := TStringList.Create;
  slMasters.Sorted := True;
  slMasters.Duplicates := dupIgnore;
  slMgfMasters := TStringList.Create;
  slDictionary := TStringList.Create;
  slDictionary.LoadFromFile(ScriptsPath + 'mp\dictionary.txt');
  slTranslations := TStringList.Create;
  slCopiedFrom := TStringList.Create;
  batch := TStringList.Create;
  OldForms := TList.Create;
  NewForms := TList.Create;
  
  // load gui elements
  gear := TPicture.Create;
  gear.LoadFromFile(ProgramPath + 'Edit Scripts\mp\assets\gear.png');
  browse := TPicture.Create;
  browse.LoadFromFile(ProgramPath + 'Edit Scripts\mp\assets\browse.png');
  
  // set up temporary directory
  try
    temp := ScriptsPath + 'mp\temp\';
    ForceDirectories(temp);
    gear.SaveToFile(temp+'test.png');
  except on x: Exception do
    AddMessage('예외: '+x.Message);
  end;
  if (not DirectoryExists(temp)) or (not FileExists(temp + 'test.png')) then begin
    AddMessage('임시 폴더 생성 실패. 스크립트가 중단됩니다.');
    SkipProcess := true;
  end;
  
  // process only file elements
  try 
    ScriptProcessElements := [etFile];
  except on Exception do
    SkipProcess := true;
  end;
end;

//=========================================================================
// Process: put files to be merged into stringlists
function Process(f: IInterface): integer;
var
  i: integer;
  s: string;
begin
  if SkipProcess then 
    exit;
    
  if (ElementType(f) = etMainRecord) then 
    exit;
    
  s := GetFileName(f);
  slSelectedFiles.AddObject(s, TObject(GetLoadOrder(f)));
end;

//=========================================================================
// Finalize: this is where all the good stuff happens
function Finalize: integer;
var
  i, j, k, rc, wait, waitTick: integer;
  f, e, group, masters, master: IInterface;
  merge, id, desc, version, fn, masterName, mergeDesc, bfn: string;
  done, b, recordFromMerge, didNothing: boolean;
  lbl: TLabel;
  pb: TProgressBar;
  today: TDateTime;
begin
  // change hint duration
  Application.HintHidePause := 10000;
  
  // check version
  try
    k := wbVersionNumber;
    version := GetVersionString(k);
    AddMessage(version);
    RemoveFilter();
  except on Exception do
    k := 0;
  end;
  // terminate script unless version is 3.0.33 or newer
  if (k < 50340096) then begin
    AddMessage('');
    EditOutOfDate('3.0.33 svn 1898', 'http://www.nexusmods.com/skyrim/mods/25859');
    FreeMemory;
    exit;
  end;
  
  MergeForm;
  
  // terminate script if mergelist contains less than one file
  if slMerge.Count < 1 then begin
    AddMessage(#13#10+'병합할 파일을 1개 이상 선택해야 합니다.  스크립트가 중단됩니다.'+#13#10);
    FreeMemory;
    exit;
  end;
  
  // terminate script if usingMO is true but moPath isn't correct
  if (usingMO) and ((moPath = '') or (moPath = '?')) then begin
    AddMessage(#13#10+
    'Mod Organizer의 경로가 유효하지 않습니다. MO 사용자가 아닐 경우, '#13#10
	'병합 고급 옵션 창에서 Mod Organizer 사용자 체크박스의 체크를 해제하세요.'#13#10);
    FreeMemory;
    exit;
  end;
  
  // provide user with asset destination helper
  if not AssetHelper then begin
    AddMessage(#13#10'병합 취소를 선택했습니다.  스크립트가 중단됩니다.'#13#10);
    FreeMemory;
    exit;
  end;
  
  // create or identify merge file
  done := False;
  mgf := nil;
  AddMessage(#13#10+'파일 병합 준비 중...');
  mgf := FileSelectM('기존 파일에 병합하거나, 새 파일을 생성하세요. '+
    #13#10#13#10+'참고: 기존 파일에 병합하는 것은 안정성이 떨어지므로' +
    #13#10+'새 파일 생성을 추천합니다.');

  // merge file confirmation or termination
  if not Assigned(mgf) then begin
    AddMessage('    병합 파일 할당 안됨.  스크립트가 중단됩니다.'+#13#10);
    FreeMemory;
    exit;
  end;
  AddMessage('    스크립트를 통해 생성될 병합 파일: ' + GetFileName(mgf) + '');
    
  // set up for saving log
  ForceDirectories(ScriptsPath + '\mp\logs');
  today := Now;
  fdt := FormatDateTime('mmddyy_hhnnss', today);
  fn := 'merge_'+fdt+'.txt';
  
  // display progress bar
  frm := TForm.Create(nil);
  try
    frm.Caption := '플러그인 병합 중...';
    frm.Width := 700;
    frm.Position := poScreenCenter;
    frm.Height := 150;
    
    lbl := TLabel.Create(frm);
    lbl.Parent := frm;
    lbl.Top := 20;
    lbl.Left := 20;
    lbl.Width := frm.Width - 55;
    lbl.Height := 30;
    lbl.Caption := '인식 중... ';
    lbl.Visible := true;
    
    pb := TProgressBar.Create(frm);
    pb.Parent := frm;
    pb.Top := 40;
    pb.Left := 20;
    pb.Width := frm.Width - 55;
    pb.Height := 20;
    pb.Step := 1;
    pb.Min := 0;
    pb.Max := 100;
    pb.Position := 0;
    
    memo := TMemo.Create(frm);
    memo.Parent := frm;
    memo.Top := 70;
    memo.Left := 20;
    memo.Width := pb.Width;
    memo.WordWrap := false;
    memo.ScrollBars := ssBoth;
    memo.Visible := false;
    memo.ReadOnly := true;
    
    btnDetails := TButton.Create(frm);
    btnDetails.Parent := frm;
    btnDetails.Top := pb.Top + pb.Height + 8;
    btnDetails.Left := pb.Left;
    btnDetails.Caption := '상세정보 보기';
    btnDetails.Width := 100;
    btnDetails.OnClick := ShowDetails;
    
    frm.Show;
    application.processmessages;
    
    // print initial log messages
    LogMessage(dashes);
    LogMessage('병합 플러그인 '+vs+': 모드 플러그인들을 병합해 줍니다.');
    LogMessage(dashes);
    LogMessage(version+#13#10);
    for i := 0 to slMerge.Count - 1 do
      LogMessage('병합: '+slMerge[i]);
    
    LogMessage(#13#10+'스크립트를 통해 생성될 병합 파일: ' + GetFileName(mgf) + '');
  
    // add masters
    lbl.Caption := '마스터 추가 중...';
    LogMessage('    병합 파일에 마스터를 추가합니다...');
    AddMastersToFile(mgf, slMasters, true);
     
    // renumber forms in files to be merged
    rCount := 0;
    if (rn = 2) then begin
      lbl.Caption := '모든 FormID 리넘버링 중...';
      RenumberAll(pb);
      LogMessage('    '+IntToStr(rCount)+' 개의 레코드가 리넘버링 되었습니다.');
    end
    else if (rn = 1) then begin
      lbl.Caption := '충돌 FormID 리넘버링 중...';
      RenumberConflicting(pb);
      LogMessage('    '+IntToStr(rCount)+' 개의 레코드가 리넘버링 되었습니다.');
    end
    else if (rn = 0) then begin
      // get NextObjectID
      NextObjectID := FindHighestFormID() + 1024;
      
      // make formID lists
      for i := 0 to slMerge.Count - 1 do begin
        f := FileByLoadOrder(Integer(slMerge.Objects[i]));
        RC := RecordCount(f) - 1;
        NewForms.Add(TStringList.Create);
        for j := 0 to RC do begin
          e := RecordByIndex(f, j);
          if (Signature(e) = 'TES4') then Continue;
          TStringList(NewForms[i]).Add(HexFormID(e));
        end;
      end;
      
      // copy File specific asets
      LogMessage(#13#10+'특수 Asset 복사 중...');
      lbl.Caption := '특수 Asset 파일을 복사합니다...';
      for i := 0 to slMerge.Count - 1 do CopyFileSpecificAssets(i);
    end;
    // set next object id
    senv(ElementByIndex(mgf, 0), 'HEDR\Next Object ID', NextObjectID);
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
    
    // mod organizer CopyGeneralAssets option
    if copyAll then begin
      didNothing := true;
      LogMessage(#13#10+'Mod Organizer 폴더에서 일반 Asset 파일을 복사합니다.');
      lbl.Caption := 'MO에서 일반 Asset 파일 복사 중...';
      for i := 0 to slMerge.Count - 1 do begin
        CopyGeneralAssets(slMerge[i]);
        Application.processmessages;
        didNothing := false;
      end;
      if didNothing then
        LogMessage('    복사된 일반 Asset가 없습니다.');
    end;
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
    
    // extract BSAs option
    if extractBSAs then begin
      didNothing := true;
      LogMessage(#13#10+'BSA 추출 중...');
      lbl.Caption := 'BSA 추출';
      for i := 0 to slMerge.Count - 1 do begin
        if (Pos('.esp', slMerge[i]) > 0) then
          bsaName := StringReplace(slMerge[i], '.esp', '.bsa', [rfReplaceAll]);
        if (Pos('.esm', slMerge[i]) > 0) then
          bsaName := StringReplace(slMerge[i], '.esm', '.bsa', [rfReplaceAll]);
        if (Pos('.bsa', bsaName) > 0) and FileExists(DataPath + bsaName) then begin
          LogMessage('    추출 중: '+bsaName+', 추출 경로: '+astPath);
          ExtractBSA(DataPath + bsaName, astPath);
          Application.processmessages;
          didNothing := false;
        end;
      end;
      if didNothing then
        LogMessage('    추출된 BSA가 없습니다.');
    end;
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);

    // copy records
    LogMessage(#13#10+'레코드 복사 중...');
    lbl.Caption := '레코드 복사 중...';
    pb.Position := 29;
    rCount := 0;
    for i := slMerge.Count - 1 downto 0 do begin
      f := FileByLoadOrder(Integer(slMerge.Objects[i]));
      LogMessage('    다음 파일에서 레코드 복사 중: '+GetFileName(f));
      if mm = 0 then MergeByRecords(f, (sp > 0)) else 
      if mm = 1 then MergeIntelligently(f, (sp > 0)) else 
      if mm = 2 then MergeByGroups(f, (sp > 0));
      pb.Position := pb.Position + 30/slMerge.Count;
      Application.processmessages;
    end;
    LogMessage('    '+IntToStr(rCount)+' 개의 레코드가 복사되었습니다.');
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
   
    // removing masters
    LogMessage(#13#10+'불필요한 마스터 제거 중...');
    lbl.Caption := '마스터 제거 중...';
    pb.Position := 59;
    masters := ElementByName(ElementByIndex(mgf, 0), 'Master Files');
    for i := ElementCount(masters) - 1 downto 0 do begin
      e := ElementByIndex(masters, i);
      masterName := GetElementNativeValues(e, 'MAST');
      if (masterName = '') then Continue;
      if slMerge.IndexOf(masterName) > -1 then begin
        LogMessage('    마스터 제거 중: '+masterName);
        RemoveElement(masters, e);
      end;
    end;
    
    // creating description
    desc := '병합에 사용된 플러그인: ';
    lbl.Caption := '병합 설명';
    pb.Position := 60;
    mergeDesc := nil;
    mergeDesc := geev(ElementByIndex(mgf, 0), 'SNAM');
    if not Assigned(mergeDesc) then begin
      Add(ElementByIndex(mgf, 0), 'SNAM', True);
      desc := '병합에 사용된 플러그인:';
    end
    else if Pos('Merged Plugin', mergeDesc) > 0 then 
      desc := mergeDesc;
    for i := 0 to slMerge.Count - 1 do begin
      mergeDesc := geev(ElementByIndex(FileByLoadOrder(Integer(slMerge.Objects[i])), 0), 'SNAM');
      if Pos('Merged Plugin', mergeDesc) > 0 then
        desc := desc+StringReplace(mergeDesc, '병합에 사용된 플러그인:', '', [rfReplaceAll])
      else
        desc := desc+#13#10+'  '+slMerge[i];
    end;
    seev(ElementByIndex(mgf, 0), 'CNAM', 'Merge Plugins Script '+vs);
    seev(ElementByIndex(mgf, 0), 'SNAM', desc);
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
    
    // second pass copying
    if (sp > 0) then begin
      // copy records again
      LogMessage(#13#10+'2차 패스 복사 실행 중...');
      pb.Position := 65;
      lbl.Caption := '레코드 복사 중 (2차 패스)...';
      rCount := 0;
      for i := slMerge.Count - 1 downto 0 do begin
        f := FileByLoadOrder(Integer(slMerge.Objects[i]));
        LogMessage('    다음 파일에서 레코드 복사 중: '+GetFileName(f));
        if (sp = 1) then begin
          if mm = 0 then MergeByRecords(f, false) else 
          if mm = 1 then MergeIntelligently(f, false) else 
          if mm = 2 then MergeByGroups(f, false);
        end
        else MergeByGroups(f);
        pb.Position := pb.Position + 30/slMerge.Count;
        Application.processmessages;
      end;
      LogMessage('    '+IntToStr(rCount)+' 개의 레코드가 복사되었습니다.');
    end;
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
    
    // create formID list group
    Add(mgf, 'FLST', true);
    // create formlists
    lbl.Caption := 'FormList 생성 중...';
    pb.Position := 98;
    LogMessage(#13#10+'FormList 생성 중...');
    for i := 0 to slMerge.Count - 1 do begin
      LogMessage('    다음 플러그인의 Formlist 생성 중: '+slMerge[i]);
      e := Add(GroupBySignature(mgf, 'FLST'), 'FLST', True);
      seev(e, 'EDID', Copy(slMerge[i], 1, Length(slMerge[i]) - 4)+'Forms');
      Add(e, 'FormIDs', True);
      try
        slev(e, 'FormIDs', TStringList(NewForms[i]));
      except on Exception do
        ; // nothing we can really do
      end;
      Application.processmessages;
    end;
    
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);

    // script is done, print confirmation messages
    pb.Position := 100;
    lbl.Caption := '플러그인이 병합되었습니다.';
    LogMessage(#13#10);
    LogMessage(dashes);
    LogMessage('플러그인 병합이 성공적으로 완료되었습니다.  레코드 수: '+IntToStr(RecordCount(mgf))+'');
    // inform user about records that failed to copy
    if (slFails.Count > 0) then begin
      ShowDetails;
      Application.processmessages;
      MessageDlg('일부 레코드 복사에 실패하여 병합된 플러그인이 완전하지 않습니다.  '	  
      '메시지 로그에 따라 해당되는 레코드를 수동으로 점검하세요.  '
      '(무시할 경우 병합된 플러그인이 제대로 작동하지 않을 수 있습니다!)', mtConfirmation, [mbOk], 0);
      LogMessage('복사 실패한 레코드: ');
      for i := 0 to slFails.Count - 1 do 
        LogMessage('    '+slFails[i]);
    end;
    LogMessage(#13#10);
  
    // save log
    memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
    
    // perform batch copy
    if batCopy and (batch.Count > 0) then begin
      ShowDetails;
      Application.processmessages;
      LogMessage('cmd 배치 복사 옵션이 활성화되어, Asset 복사시 cmd 윈도우창이 실행됩니다.');
      LogMessage('복사 도중 ''Exit''를 눌러 종료하거나, 중간에 창을 닫지 마세요.');
      LogMessage('복사가 완료되면 cmd 창이 자동으로 닫힙니다.');
      LogMessage(#13#10#13#10);
      memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
      bfn := temp+'merge_'+fdt+'.bat';
      batch.SaveToFile(bfn);
      ShellExecute(TForm(frmMain).Handle, 'open', bfn, '', ExtractFilePath(bfn), SW_SHOWNORMAL);
    end;
    
    // wait for user to close form if details visible
    if (memo.Visible) then begin
      frm.Visible := false;
      frm.ShowModal;
    end;
  except on x : Exception do begin
      // merge failed
      AddMessage(#13#10'병합 실패.  예외: '+x.Message);
      LogMessage(#13#10'병합 실패.  예외: '+x.Message);
      memo.Lines.SaveToFile(ScriptsPath+'\mp\logs\'+fn);
      pb.Position := 0;
      lbl.Caption := '병합 실패.  예외: '+x.Message;
      if not memo.Visible then ShowDetails;
      frm.Visible := false;
      frm.ShowModal;
      Application.processmessages;
    end;
  end;
  // free form
  frm.Free;
  // free memory
  FreeMemory;
  // clear temp folder if it's not = to TempPath
  if deleteTemp then begin
    if not DeleteDirectory(temp, true) then begin
      AddMessage(#13#10'임시 폴더 삭제에 실패했습니다.');
      AddMessage('병합 스크립트가 끝난 뒤 다음 폴더를 삭제하세요: '+temp);
    end;
  end;
  // return hinthidepasue to default value
  Application.HintHidePause := 1000;
  // call RemoveFilter() to update xEdit GUI
  try
    RemoveFilter();
  except on Exception do begin
      AddMessage('');
      EditOutOfDate('3.0.33 svn 1898', 'http://www.nexusmods.com/skyrim/mods/25859');
    end;
  end;
end;

end.